#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Roulette Wheel Selection implementation. '''

from random import random
from bisect import bisect_right
from itertools import accumulate

from ...plugin_interfaces.operators.selection import Selection

class RouletteWheelSelection(Selection):
    ''' Selection operator with fitness proportionate selection(FPS) or
    so-called roulette-wheel selection implementation.
    '''
    def __init__(self):
        pass

    def select(self, population, fitness):
        ''' Select a pair of parent using FPS algorithm.

        :param population: Population where the selection operation occurs.
        :type population: :obj:`lsga.components.Population`

        :return: Selected parents (a father and a mother)
        :rtype: list of :obj:`lsga.components.IndividualBase`
        '''
        # Normalize fitness values for all individuals.
        fit = population.all_fits(fitness)
        min_fit = min(fit)
        fit = [(i - min_fit) for i in fit]

        # Create roulette wheel.
        sum_fit = sum(fit)
        wheel = list(accumulate([i/sum_fit for i in fit]))

        # Select a father and a mother.
        father_idx = bisect_right(wheel, random())
        father = population[father_idx]
        mother_idx = (father_idx + 1) % len(wheel)
        mother = population[mother_idx]

        return father, mother




#!/usr/bin/env python
# -*- coding: utf-8 -*-
''' Definition of individual class with decimal encoding.
'''

from .individual import IndividualBase


class DecimalIndividual(IndividualBase):
    ''' Individual with decimal encoding.

    :param ranges: value ranges for all entries in solution.
    :type ranges: tuple list

    :param eps: decrete precisions for binary encoding, default is 0.001.
    :type eps: float or float list (with the same length with ranges)
    '''
    def __init__(self, ranges, eps=0.001):
        super(self.__class__, self).__init__(ranges, eps)
        # Initialize it randomly.
        self.init()

    def encode(self):
        ''' Encode solution to gene sequence
        '''
        return self.solution

    def decode(self):
        ''' Decode gene sequence to decimal solution
        '''
        return self.solution




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from random import uniform
from copy import deepcopy


class SolutionRanges(object):
    ''' Descriptor for solution ranges.
    '''
    def __init__(self):
        self.__ranges = []

    def __get__(self, obj, owner):
        return self.__ranges

    def __set__(self, obj, ranges):
        # Check.
        if type(ranges) not in [tuple, list]:
            raise TypeError('solution ranges must be a list of range tuples')
        for rng in ranges:
            if type(rng) not in [tuple, list]:
                raise TypeError('range({}) is not a tuple containing two numbers'.format(rng))
            if len(rng) != 2:
                raise ValueError('length of range({}) not equal to 2')
            a, b = rng
            if a >= b:
                raise ValueError('Wrong range value {}'.format(rng))
        # Assignment.
        self.__ranges = ranges


class DecretePrecision(object):
    ''' Descriptor for individual decrete precisions.
    '''
    def __init__(self):
        self.__precisions = []

    def __get__(self, obj, owner):
        return self.__precisions

    def __set__(self, obj, precisions):
        if type(precisions) in [int, float]:
            precisions = [precisions]*len(obj.ranges)
        # Check.
        if type(precisions) not in [tuple, list]:
            raise TypeError('precisions must be a list of numbers')
        if len(precisions) != len(obj.ranges):
            raise ValueError('Lengths of eps and ranges should be the same')
        for (a, b), eps in zip(obj.ranges, precisions):
            if eps > (b - a):
                msg = 'Invalid precision {} in range ({}, {})'.format(eps, a, b)
                raise ValueError(msg)
        self.__precisions = precisions


class IndividualBase(object):
    ''' Base class for individuals.

    :param ranges: value ranges for all entries in solution.
    :type ranges: tuple list

    :param eps: decrete precisions for binary encoding, default is 0.001.
    :type eps: float or float list (with the same length with ranges)
    '''
    # Solution ranges.
    ranges = SolutionRanges()

    # Orginal decrete precisions (provided by users).
    eps = DecretePrecision()
    
    # Actual decrete precisions used in GA.
    precisions = DecretePrecision()

    def __init__(self, ranges, eps):
        self.ranges = ranges
        self.eps = eps
        self.precisions = eps

        self.solution, self.chromsome = [], []

    def init(self, chromsome=None, solution=None):
        ''' Initialize the individual by providing chromsome or solution.

        :param chromsome: chromesome sequence for the individual
        :type chromsome: list of (float / int)

        :param solution: the variable vector of the target function.
        :type solution: list of float

        .. Note::
            If both chromsome and solution are provided, only the chromsome would
            be used. If neither is provided, individual would be initialized randomly.
        '''
        if not any([chromsome, solution]):
            self.solution = self._rand_solution()
            self.chromsome = self.encode()
        elif chromsome:
            self.chromsome = chromsome
            self.solution = self.decode()
        else:
            self.solution = solution
            self.chromsome = self.encode()

        return self

    def clone(self):
        ''' Clone a new individual from current one.
        '''
        indv = self.__class__(deepcopy(self.ranges), eps=deepcopy(self.eps))
        indv.init(chromsome=deepcopy(self.chromsome))
        return indv


    def encode(self):
        ''' **NEED IMPLIMENTATION**

        Convert solution to chromsome sequence.

        :return: The chromsome sequence
        :rtype: list of float
        '''
        raise NotImplementedError

    def decode(self):
        ''' **NEED IMPLIMENTATION**

        Convert chromsome sequence to solution.

        :return: The solution vector
        :rtype: list of float
        '''
        raise NotImplementedError

    def _rand_solution(self):
        ''' Initialize individual solution randomly.
        '''
        solution = []
        for eps, (a, b) in zip(self.precisions, self.ranges):
            n_intervals = (b - a)//eps
            n = int(uniform(0, n_intervals + 1))
            solution.append(a + n*eps)
        return solution





#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Test the dynmaic linear scaling decorator of GA engine.
'''

import unittest
from math import sin, cos

from .. import GAEngine
from ..components import BinaryIndividual
from ..components import Population
from ..operators import RouletteWheelSelection
from ..operators import UniformCrossover
from ..operators import FlipBitMutation


class DynamicLinearScalingTest(unittest.TestCase):

    def setUp(self):
        self.maxDiff = True

    def test_run(self):
        '''
        Make sure GA engine can run correctly.
        '''
        indv_template = BinaryIndividual(ranges=[(0, 10)], eps=0.001)
        population = Population(indv_template=indv_template, size=50).init()

        # Create genetic operators.
        selection = RouletteWheelSelection()
        crossover = UniformCrossover(pc=0.8, pe=0.5)
        mutation = FlipBitMutation(pm=0.1)

        # Create genetic algorithm engine.
        engine = GAEngine(population=population, selection=selection,
                          crossover=crossover, mutation=mutation)

        @engine.fitness_register
        @engine.dynamic_linear_scaling()
        def fitness(indv):
            x, = indv.solution
            return x + 10*sin(5*x) + 7*cos(4*x)

        engine.run(50)

if '__main__' == __name__:
    suite = unittest.TestLoader().loadTestsFromTestCase(DynamicLinearScalingTest)
    unittest.TextTestRunner(verbosity=2).run(suite)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Test case for Individual.
'''

import unittest

from ..components import BinaryIndividual

class IndividualTest(unittest.TestCase):

    def setUp(self):
        self.maxDiff = True

    def test_binary_encoding(self):
        ''' Make sure individual can decode and encode binary gene correctly.
        '''
        indv = BinaryIndividual(ranges=[(0, 1)], eps=0.001)
        indv.init(solution=[0.398])

        # Test binary chromsome.
        ref_chromsome = [0, 1, 1, 0, 0, 1, 0, 1, 1]
        self.assertListEqual(indv.chromsome, ref_chromsome)

        # Test decode.
        self.assertListEqual(indv.decode(), [0.396484375])

        indv = BinaryIndividual(ranges=[(0, 1), (-1, 1)],  eps=0.001)
        indv.init(solution=[0.398, 0.66])

        # Test binary chromsome.
        ref_chromsome = [0, 1, 1, 0, 0, 1, 0, 1, 1, 1, 1, 0, 1, 0, 1, 0, 0, 0, 1]
        self.assertListEqual(indv.chromsome, ref_chromsome)

        # Test decode.
        self.assertListEqual(indv.decode(), [0.396484375, 0.658203125])

    def test_descriptors(self):
        ''' Make sure descriptors can check the parameters correctly.
        '''
        self.assertRaises(TypeError, BinaryIndividual, ranges=0.1, eps=0.001)
        self.assertRaises(TypeError, BinaryIndividual, ranges=[(0, 1)], eps='asdf')
        self.assertRaises(ValueError, BinaryIndividual, ranges=[(0, 1)], eps=10.0)
        self.assertRaises(ValueError, BinaryIndividual, ranges=[(0, 1)], eps=[1, 2])

    def test_init(self):
        ''' Make sure the individual can be initialized correctly.
        '''
        indv = BinaryIndividual(ranges=[(0, 1)], eps=0.001)

        # Check chromsome initialization.
        indv.init(chromsome=[0, 1, 1, 0, 0, 0, 1, 1, 1, 0])
        
        self.assertListEqual([0, 1, 1, 0, 0, 0, 1, 1, 1, 0], indv.chromsome)
        self.assertListEqual(indv.solution, [0.388671875])

        # Check solution initialization.
        indv.init(solution=[0.398])
        
        self.assertListEqual(indv.solution, [0.398])
        self.assertListEqual(indv.chromsome, [0, 1, 1, 0, 0, 1, 0, 1, 1])

    def test_clone(self):
        ''' Make sure individual can be cloned correctly.
        '''
        indv = BinaryIndividual(ranges=[(0, 1)],
                                eps=0.001).init(solution=[0.398])
        indv_clone = indv.clone()

        self.assertListEqual(indv.chromsome, indv_clone.chromsome)
        self.assertAlmostEqual(indv.solution[0], indv_clone.solution[0], places=2)
        self.assertEqual(indv.ranges, indv_clone.ranges)
        self.assertEqual(indv.eps, indv_clone.eps)

    def test_multi_precisions(self):
        ''' Make sure we can construct individual using different decrete precisions.
        '''
        indv = BinaryIndividual(ranges=[(0, 1), (0, 10)], eps=[0.01, 1.0]).init(solution=[0.3, 0.5])
        self.assertNotEqual(indv.precisions[0], indv.precisions[1])

if '__main__' == __name__:
    suite = unittest.TestLoader().loadTestsFromTestCase(IndividualTest)
    unittest.TextTestRunner(verbosity=2).run(suite)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Test case for built-in Flip Bit mutation operator.
'''

import unittest

from ..components import BinaryIndividual, DecimalIndividual
from ..operators.mutation.flip_bit_mutation import FlipBitMutation

class FlipBitMutationTest(unittest.TestCase):

    def setUp(self):
        self.maxDiff

    def test_mutate_binary_indv(self):
        ''' Make sure the individual can be mutated correctly.
        '''
        indv = BinaryIndividual(ranges=[(0, 1)]).init(solution=[0.398])
        mutation = FlipBitMutation(pm=1.0)
        chromsome_before = [0, 1, 1, 0, 0, 1, 0, 1, 1]
        chromsome_after = [1, 0, 0, 1, 1, 0, 1, 0, 0]
        self.assertListEqual(indv.chromsome, chromsome_before)
        mutation.mutate(indv, engine=None)
        self.assertListEqual(indv.chromsome, chromsome_after)

    def test_mutate_decimal_indv(self):
        ''' Make sure individual with decimal encoding can be mutated correctly.
        '''
        indv = DecimalIndividual(ranges=[(0, 1), (0, 2)]).init(solution=[0.5, 1.5])
        mutation = FlipBitMutation(pm=1.0)
        chromsome_before = [0.5, 1.5]
        self.assertListEqual(indv.chromsome, chromsome_before)
        mutation.mutate(indv, engine=None)
        for a, b in zip(indv.chromsome, chromsome_before):
            self.assertNotEqual(a, b)

if '__main__' == __name__:
    suite = unittest.TestLoader().loadTestsFromTestCase(FlipBitMutationTest)
    unittest.TextTestRunner(verbosity=2).run(suite)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .metaclasses import AnalysisMeta


class OnTheFlyAnalysis(metaclass=AnalysisMeta):
    ''' Class for providing an interface to easily extend and customize the behavior
    of the on-the-fly analysis functionality of lsga.

    Attribute:

        interval(:obj:`int`): The analysis interval in evolution iteration, default 
                              value is 1 meaning analyze every step.

        master_only(:obj:`bool`): Flag for if the analysis plugin is only effective 
                                  in master process. Default is True.
    '''
    # Only used in master process?
    master_only = False

    # Analysis interval.
    interval = 1

    def setup(self, ng, engine):
        ''' Function called right before the start of genetic algorithm main iteration
        to allow for custom setup of the analysis object.

        :param ng: The number of generation.
        :type ng: int

        :param engine: The current GAEngine where the analysis is running.
        :type engine: gaft.engine.GAEngine
        '''
        raise NotImplementedError

    def register_step(self, g, population, engine):
        '''
        Function called in each iteration step.

        :param g: Current generation number.
        :type g: int

        :param population: The up to date population of the iteration.
        :type population: Population

        :param engine: The current GAEngine where the analysis is running.
        :type engine: gaft.engine.GAEngine
        '''
        raise NotImplementedError

    def finalize(self, population, engine):
        '''
        Called after the iteration to allow for custom finalization and
        post-processing of the collected data.

        :param population: The up to date population of the iteration.
        :type population: Population

        :param engine: The current GAEngine where the analysis is running.
        :type engine: gaft.engine.GAEngine
        '''
        raise NotImplementedError




#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Module for Individual with binary encoding.
'''
from math import log2
from itertools import accumulate
import logging

from .individual import IndividualBase
from ..mpiutil import MPIUtil

mpi = MPIUtil()


class BinaryIndividual(IndividualBase):
    '''
    Class for individual in population. Random solution will be initialized
    by default.

    :param ranges: value ranges for all entries in solution.
    :type ranges: tuple list

    :param eps: decrete precisions for binary encoding, default is 0.001.
    :type eps: float or float list (with the same length with ranges)

    .. Note:

        The decrete precisions for different components in varants may be
        adjusted automatically (possible precision loss) if eps and ranges
        are not appropriate.
    '''
    def __init__(self, ranges, eps=0.001):
        super(self.__class__, self).__init__(ranges, eps)

        # Lengths for all binary sequence in chromsome and adjusted decrete precisions.
        self.lengths = []

        for i, ((a, b), eps) in enumerate(zip(self.ranges, self.eps)):
            length = int(log2((b - a)/eps))
            precision = (b - a)/(2**length)
            self.lengths.append(length)
            self.precisions[i] = precision

        # The start and end indices for each gene segment for entries in solution.
        self.gene_indices = self._get_gene_indices()

        # Initialize individual randomly.
        self.init()

    def encode(self):
        ''' Encode solution to gene sequence in individual using different encoding.
        '''
        chromsome = []
        for var, (a, _), length, eps in zip(self.solution, self.ranges,
                                            self.lengths, self.precisions):
            chromsome.extend(self.binarize(var-a, eps, length))

        return chromsome

    def decode(self):
        ''' Decode gene sequence to solution of target function.
        '''
        solution =  [self.decimalize(self.chromsome[start: end], eps, lower_bound)
                     for (start, end), (lower_bound, _), eps in
                     zip(self.gene_indices, self.ranges, self.precisions)]
        return solution

    def _get_gene_indices(self):
        '''
        Helper function to get gene slice indices.
        '''
        end_indices = list(accumulate(self.lengths))
        start_indices = [0] + end_indices[: -1]
        return list(zip(start_indices, end_indices))

    @staticmethod
    def binarize(decimal, eps, length):
        ''' Helper function to convert a float to binary sequence.

        :param decimal: the decimal number to be converted
        :type decimal: float

        :param eps: the decrete precision of binary sequence
        :type eps: float

        :param length: the length of binary sequence.
        :type length: int
        '''
        n = int(decimal/eps)
        bin_str = '{:0>{}b}'.format(n, length)
        return [int(i) for i in bin_str]

    @staticmethod
    def decimalize(binary, eps, lower_bound):
        ''' Helper function to convert a binary sequence back to decimal number.

        :param binary: The binary list to be converted
        :type binary: list of int

        :param eps: the decrete precision of binary sequence
        :type eps: float

        :param lower_bound: the lower bound for decimal number
        :type lower_bound: float
        '''
        bin_str = ''.join([str(bit) for bit in binary])
        return lower_bound + int(bin_str, 2)*eps




#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Test case for built-in Roulette Wheel selection
'''

import unittest

from lsga.components import Population, BinaryIndividual
from lsga.operators.selection.roulette_wheel_selection import RouletteWheelSelection

class RouletteWheelSelectionTest(unittest.TestCase):

    def setUp(self):
        self.maxDiff
        def fitness(indv):
            x, = indv.solution
            return x**3 - 60*x**2 + 900*x + 100
        self.fitness = fitness

    def test_selection(self):
        indv = BinaryIndividual(ranges=[(0, 30)])
        p = Population(indv)
        p.init()

        selection = RouletteWheelSelection()
        father, mother = selection.select(p, fitness=self.fitness)

        self.assertTrue(isinstance(father, BinaryIndividual))
        self.assertTrue(isinstance(mother, BinaryIndividual))
        self.assertNotEqual(father.chromsome, mother.chromsome)

if '__main__' == __name__:
    suite = unittest.TestLoader().loadTestsFromTestCase(RouletteWheelSelectionTest)
    unittest.TextTestRunner(verbosity=2).run(suite)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Test case for engine run.
'''

import unittest
from math import sin, cos

from .. import GAEngine
from ..components import BinaryIndividual, DecimalIndividual
from ..components import Population
from ..operators import RouletteWheelSelection
from ..operators import UniformCrossover
from ..operators import FlipBitMutation


class GAEngineTest(unittest.TestCase):

    def setUp(self):
        self.maxDiff = True

    def test_binary_run(self):
        '''
        Make sure GA engine can run correctly.
        '''
        indv_template = BinaryIndividual(ranges=[(0, 10)], eps=0.001)
        population = Population(indv_template=indv_template, size=50).init()

        # Create genetic operators.
        selection = RouletteWheelSelection()
        crossover = UniformCrossover(pc=0.8, pe=0.5)
        mutation = FlipBitMutation(pm=0.1)

        # Create genetic algorithm engine.
        engine = GAEngine(population=population, selection=selection,
                          crossover=crossover, mutation=mutation)

        @engine.fitness_register
        @engine.minimize
        def fitness(indv):
            x, = indv.solution
            return x + 10*sin(5*x) + 7*cos(4*x)

        engine.run(50)

    def test_decimal_run(self):
        '''
        Make sure GA engine can run correctly with decimal encoding individuals.
        '''
        indv_template = DecimalIndividual(ranges=[(0, 10)], eps=0.001)
        population = Population(indv_template=indv_template, size=50).init()

        # Create genetic operators.
        selection = RouletteWheelSelection()
        crossover = UniformCrossover(pc=0.8, pe=0.5)
        mutation = FlipBitMutation(pm=0.1)

        # Create genetic algorithm engine.
        engine = GAEngine(population=population, selection=selection,
                          crossover=crossover, mutation=mutation)

        @engine.fitness_register
        @engine.minimize
        def fitness(indv):
            x, = indv.solution
            return x + 10*sin(5*x) + 7*cos(4*x)

        engine.run(50)

if '__main__' == __name__:
    suite = unittest.TestLoader().loadTestsFromTestCase(GAEngineTest)
    unittest.TextTestRunner(verbosity=2).run(suite)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Test case for MPI utility class.
'''

import unittest

from lsga.mpiutil import MPIUtil

mpi = MPIUtil()


class MPIUtilTest(unittest.TestCase):

    def setUp(self):
        self.maxDiff = True

    def test_split_seq(self):
        '''
        Make sure sequence can be splited according to process number correctly.
        '''
        seq = list(range(10))

        if mpi.size == 1:
            self.assertListEqual(mpi.split_seq(seq), seq)
        elif mpi.size == 2:
            if mpi.rank == 0:
                self.assertListEqual(mpi.split_seq(seq), [0, 1, 2, 3, 4])
            elif mpi.rank == 1:
                self.assertListEqual(mpi.split_seq(seq), [5, 6, 7, 8, 9])

    def test_merge_seq(self):
        '''
        Make sure sequence in different processes can be merged correctly.
        '''
        if mpi.size == 1:
            self.assertListEqual(mpi.merge_seq([1, 2, 3]), [1, 2, 3])
        if mpi.size == 2:
            if mpi.rank == 0:
                recv_data = mpi.merge_seq([1, 2, 3])
            elif mpi.rank == 1:
                recv_data = mpi.merge_seq([2, 3, 4])
            self.assertListEqual(recv_data, [1, 2, 3, 2, 3, 4])

    def test_split_size(self):
        '''
        Make sure a size number can be splited correctly.
        '''
        if mpi.size == 1:
            self.assertEqual(mpi.split_size(50), 50)
        if mpi.size == 2:
            if mpi.rank == 0:
                self.assertEqual(mpi.split_size(49), 25)
                self.assertEqual(mpi.split_size(1), 1)
            elif mpi.rank == 1:
                self.assertEqual(mpi.split_size(49), 24)
                self.assertEqual(mpi.split_size(1), 0)

if '__main__' == __name__:
    suite = unittest.TestLoader().loadTestsFromTestCase(MPIUtilTest)
    unittest.TextTestRunner(verbosity=2).run(suite)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Test case for built-in Uniform Crossover operator.
'''

import unittest

from lsga.components import BinaryIndividual
from lsga.operators.crossover.uniform_crossover import UniformCrossover

class UniformCrossoverTest(unittest.TestCase):

    def setUp(self):
        self.maxDiff

    def test_cross(self):
        ''' Make sure individuals can be crossed correctly.
        '''
        father = BinaryIndividual(ranges=[(0, 1)]).init(solution=[0.398])
        mother = BinaryIndividual(ranges=[(0, 1)]).init(solution=[0.298])
        crossover = UniformCrossover(pc=1.0, pe=0.5)
        child1, child2 = crossover.cross(father, mother)

if '__main__' == __name__:
    suite = unittest.TestLoader().loadTestsFromTestCase(UniformCrossoverTest)
    unittest.TextTestRunner(verbosity=2).run(suite)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Uniform Crossover operator implementation. '''

from random import random
from copy import deepcopy

from ...plugin_interfaces.operators.crossover import Crossover


class UniformCrossover(Crossover):
    ''' Crossover operator with uniform crossover algorithm,
    see https://en.wikipedia.org/wiki/Crossover_(genetic_algorithm)

    :param pc: The probability of crossover (usaully between 0.25 ~ 1.0)
    :type pc: float in (0.0, 1.0]

    :param pe: Gene exchange probability.
    :type pe: float in range (0.0, 1.0]
    '''
    def __init__(self, pc, pe=0.5):
        if pc <= 0.0 or pc > 1.0:
            raise ValueError('Invalid crossover probability')
        self.pc = pc

        if pe <= 0.0 or pe > 1.0:
            raise ValueError('Invalid genome exchange probability')
        self.pe = pe

    def cross(self, father, mother):
        ''' Cross chromsomes of parent using uniform crossover method.

        :param population: Population where the selection operation occurs.
        :type population: :obj:`lsga.components.Population`

        :return: Selected parents (a father and a mother)
        :rtype: list of :obj:`lsga.components.IndividualBase`
        '''
        do_cross = True if random() <= self.pc else False

        if not do_cross:
            return father.clone(), mother.clone()

        # Chromsomes for two children.
        chrom1 = deepcopy(father.chromsome)
        chrom2 = deepcopy(mother.chromsome)

        for i, (g1, g2) in enumerate(zip(chrom1, chrom2)):
            do_exchange = True if random() < self.pe else False
            if do_exchange:
                chrom1[i], chrom2[i] = g2, g1

        child1, child2 = father.clone(), father.clone()
        child1.init(chromsome=chrom1)
        child2.init(chromsome=chrom2)

        return child1, child2




#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Package for built-in genetic operators '''

from .crossover.uniform_crossover import UniformCrossover
from .selection.roulette_wheel_selection import RouletteWheelSelection
from .selection.tournament_selection import TournamentSelection
from .selection.linear_ranking_selection import LinearRankingSelection
from .selection.exponential_ranking_selection import ExponentialRankingSelection
from .mutation.flip_bit_mutation import FlipBitMutation
from .mutation.flip_bit_mutation import FlipBitBigMutation




#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Genetic Algorithm engine definition
'''

import logging
import math
from functools import wraps

# Imports for profiling.
import cProfile
import pstats
import os

from .components import IndividualBase, Population
from .plugin_interfaces.operators import Selection, Crossover, Mutation
from .plugin_interfaces.analysis import OnTheFlyAnalysis
from .mpiutil import MPIUtil

mpi = MPIUtil()

def do_profile(filename, sortby='tottime'):
    ''' Constructor for function profiling decorator.
    '''
    def _do_profile(func):
        ''' Function profiling decorator.
        '''
        @wraps(func)
        def profiled_func(*args, **kwargs):
            '''
            Decorated function.
            '''
            # Flag for doing profiling or not.
            DO_PROF = os.getenv('PROFILING')

            if DO_PROF:
                profile = cProfile.Profile()
                profile.enable()
                result = func(*args, **kwargs)
                profile.disable()
                ps = pstats.Stats(profile).sort_stats(sortby)
                ps.dump_stats(filename)
            else:
                result = func(*args, **kwargs)
            return result
        return profiled_func

    return _do_profile


class StatVar(object):
    def __init__(self, name):
        ''' Descriptor for statistical variables which need to be memoized when
        engine running.
        '''
        # Protected.
        self.name = '_{}'.format(name)

    def __get__(self, engine, cls):
        '''
        Getter.
        '''
        stat_var = getattr(engine, self.name)
        if stat_var is None:
            if 'min' in self.name and 'ori' in self.name:
                stat_var = engine.population.min(engine.ori_fitness)
            elif 'min' in self.name:
                stat_var = engine.population.min(engine.fitness)
            elif 'max' in self.name and 'ori' in self.name:
                stat_var = engine.population.max(engine.ori_fitness)
            elif 'max' in self.name:
                stat_var = engine.population.max(engine.fitness)
            elif 'mean' in self.name and 'ori' in self.name:
                stat_var = engine.population.mean(engine.ori_fitness)
            elif 'mean' in self.name:
                stat_var = engine.population.mean(engine.fitness)
            setattr(engine, self.name, stat_var)
        return stat_var

    def __set__(self, engine, value):
        '''
        Setter.
        '''
        setattr(engine, self.name, value)


class GAEngine(object):
    ''' Class for representing a Genetic Algorithm engine. The class is the 
    central object in GAFT framework for running a genetic algorithm optimization.
    Once the population with individuals,  a set of genetic operators and fitness 
    function are setup, the engine object unites these informations and provide 
    means for running a genetic algorthm optimization.

    :param population: The Population to be reproduced in evolution iteration.
    :type population: :obj:`lsga.components.Population`

    :param selection: The Selection to be used for individual seleciton.
    :type selection: :obj:`lsga.plugin_interfaces.operators.Selection`

    :param crossover: The Crossover to be used for individual crossover.
    :type crossover: :obj:`lsga.plugin_interfaces.operators.Crossover`

    :param mutation: The Mutation to be used for individual mutation.
    :type mutation: :obj:`lsga.plugin_interfaces.operators.Mutation`

    :param fitness: The fitness calculation function for an individual in population.
    :type fitness: function

    :param analysis: All analysis class for on-the-fly analysis.
    :type analysis: :obj:`OnTheFlyAnalysis` list
    '''
    # Statistical attributes for population.
    fmax, fmin, fmean = StatVar('fmax'), StatVar('fmin'), StatVar('fmean')
    ori_fmax, ori_fmin, ori_fmean = (StatVar('ori_fmax'),
                                     StatVar('ori_fmin'),
                                     StatVar('ori_fmean'))

    def __init__(self, population, selection, crossover, mutation,
                 fitness=None, analysis=None):
        # Set logger.
        logger_name = 'lsga.{}'.format(self.__class__.__name__)
        self.logger = logging.getLogger(logger_name)

        # Attributes assignment.
        self.population = population
        self.fitness = fitness
        self.selection = selection
        self.crossover = crossover
        self.mutation = mutation
        self.analysis = [] if analysis is None else [a() for a in analysis]

        # Maxima and minima in population.
        self._fmax, self._fmin, self._fmean = None, None, None
        self._ori_fmax, self._ori_fmin, self._ori_fmean = None, None, None

        # Default fitness functions.
        self.ori_fitness = None if self.fitness is None else self.fitness

        # Store current generation number.
        self.current_generation = -1  # Starts from 0.

        # Check parameters validity.
        self._check_parameters()

    @do_profile(filename='gaft_run.prof')
    def run(self, ng=10):
        ''' Run the Genetic Algorithm optimization iteration with specified parameters.

        :param ng: Evolution iteration steps (generation number)
        :type ng: int
        '''
        if self.fitness is None:
            raise AttributeError('No fitness function in GA engine')

        self._update_statvars()

        # Setup analysis objects.
        for a in self.analysis:
            a.setup(ng=ng, engine=self)

        # Enter evolution iteration.
        try:
            for g in range(ng):
                self.current_generation = g

                # The best individual in current population. 
                if mpi.is_master:
                    best_indv = self.population.best_indv(self.fitness)
                else:
                    best_indv = None
                best_indv = mpi.bcast(best_indv)

                # Scatter jobs to all processes.
                local_indvs = []
                # NOTE: One series of genetic operation generates 2 new individuals.
                local_size = mpi.split_size(self.population.size // 2)

                # Fill the new population.
                for _ in range(local_size):
                    # Select father and mother.
                    parents = self.selection.select(self.population, fitness=self.fitness)
                    # Crossover.
                    children = self.crossover.cross(*parents)
                    # Mutation.
                    children = [self.mutation.mutate(child, self) for child in children]
                    # Collect children.
                    local_indvs.extend(children)

                # Gather individuals from all processes.
                indvs = mpi.merge_seq(local_indvs)
                # Retain the previous best individual.
                indvs[0] = best_indv
                # The next generation.
                self.population.individuals = indvs

                # Update statistic variables.
                self._update_statvars()

                # Run all analysis if needed.
                for a in self.analysis:
                    if g % a.interval == 0:
                        a.register_step(g=g, population=self.population, engine=self)
        except Exception as e:
            # Log exception info.
            if mpi.is_master:
                msg = '{} exception is catched'.format(type(e).__name__)
                self.logger.exception(msg)
            raise e
        finally:
            # Recover current generation number.
            self.current_generation = -1
            # Perform the analysis post processing.
            for a in self.analysis:
                a.finalize(population=self.population, engine=self)

    def _update_statvars(self):
        '''
        Private helper function to update statistic variables in GA engine, like
        maximum, minimum and mean values.
        '''
        # Wrt original fitness.
        self.ori_fmax = self.population.max(self.ori_fitness)
        self.ori_fmin = self.population.min(self.ori_fitness)
        self.ori_fmean = self.population.mean(self.ori_fitness)

        # Wrt decorated fitness.
        self.fmax = self.population.max(self.fitness)
        self.fmin = self.population.min(self.fitness)
        self.fmean = self.population.mean(self.fitness)

    def _check_parameters(self):
        '''
        Helper function to check parameters of engine.
        '''
        if not isinstance(self.population, Population):
            raise TypeError('population must be a Population object')
        if not isinstance(self.selection, Selection):
            raise TypeError('selection operator must be a Selection instance')
        if not isinstance(self.crossover, Crossover):
            raise TypeError('crossover operator must be a Crossover instance')
        if not isinstance(self.mutation, Mutation):
            raise TypeError('mutation operator must be a Mutation instance')

        for ap in self.analysis:
            if not isinstance(ap, OnTheFlyAnalysis):
                msg = '{} is not subclass of OnTheFlyAnalysis'.format(ap.__name__)
                raise TypeError(msg)

    # Decorators.

    def fitness_register(self, fn):
        ''' A decorator for fitness function register.

        :param fn: Fitness function to be registered
        :type fn: function
        '''
        @wraps(fn)
        def _fn_with_fitness_check(indv):
            '''
            A wrapper function for fitness function with fitness value check.
            '''
            # Check indv type.
            if not isinstance(indv, IndividualBase):
                raise TypeError('indv\'s class must be subclass of IndividualBase')

            # Check fitness.
            fitness = fn(indv)
            is_invalid = (type(fitness) is not float) or (math.isnan(fitness))
            if is_invalid:
                msg = 'Fitness value(value: {}, type: {}) is invalid'
                msg = msg.format(fitness, type(fitness))
                raise ValueError(msg)
            return fitness

        self.fitness = _fn_with_fitness_check
        if self.ori_fitness is None:
            self.ori_fitness = _fn_with_fitness_check

    def analysis_register(self, analysis_cls):
        ''' A decorator for analysis regsiter.

        :param analysis_cls: The analysis to be registered
        :type analysis_cls: :obj:`lsga.plugin_interfaces.OnTheFlyAnalysis`
        '''
        if not issubclass(analysis_cls, OnTheFlyAnalysis):
            raise TypeError('analysis class must be subclass of OnTheFlyAnalysis')

        # Add analysis instance to engine.
        analysis = analysis_cls()
        self.analysis.append(analysis)

    # Functions for fitness scaling.

    def linear_scaling(self, target='max', ksi=0.5):
        '''
        A decorator constructor for fitness function linear scaling.

        :param target: The optimization target, maximization or minimization,
                       possible value: 'max', 'min'
        :type target: str

        :param ksi: Selective pressure adjustment value.
        :type ksi: float

        .. Note::

            Linear Scaling:
                1. :math:`arg \max f(x)`, then the scaled fitness would be :math:`f - \min f(x) + {\\xi}`
                2. :math:`arg \min f(x)`, then the scaled fitness would be :math:`\max f(x) - f(x) + {\\xi}`

        '''
        def _linear_scaling(fn):
            # For original fitness calculation.
            self.ori_fitness = fn

            @wraps(fn)
            def _fn_with_linear_scaling(indv):
                # Original fitness value.
                f = fn(indv)

                # Determine the value of a and b.
                if target == 'max':
                    f_prime = f - self.ori_fmin + ksi
                elif target == 'min':
                    f_prime = self.ori_fmax - f + ksi
                else:
                    raise ValueError('Invalid target type({})'.format(target))
                return f_prime

            return _fn_with_linear_scaling

        return _linear_scaling

    def dynamic_linear_scaling(self, target='max', ksi0=2, r=0.9):
        '''
        A decorator constructor for fitness dynamic linear scaling.

        :param target: The optimization target, maximization or minimization
                       possible value: 'min' or 'max'
        :type target: str

        :param ksi0: Initial selective pressure adjustment value, default value is 2
        :type ksi0: float

        :param r: The reduction factor for selective pressure adjustment value,
                  ksi^(k-1)*r is the adjustment value for generation k, default
                  value is 0.9
        :type r: float in range [0.9, 0.999]

        .. Note::
            Dynamic Linear Scaling:

            For maximizaiton, :math:`f' = f(x) - \min f(x) + {\\xi}^{k}`, :math:`k` is generation number.
        '''
        def _dynamic_linear_scaling(fn):
            # For original fitness calculation.
            self.ori_fitness = fn

            @wraps(fn)
            def _fn_with_dynamic_linear_scaling(indv):
                f = fn(indv)
                k = self.current_generation + 1

                if target == 'max':
                    f_prime = f - self.ori_fmin + ksi0*(r**k)
                elif target == 'min':
                    f_prime = self.ori_fmax - f + ksi0*(r**k)
                else:
                    raise ValueError('Invalid target type({})'.format(target))
                return f_prime

            return _fn_with_dynamic_linear_scaling

        return _dynamic_linear_scaling

    def minimize(self, fn):
        ''' A decorator for minimizing the fitness function.

        :param fn: Original fitness function
        :type fn: function
        '''
        @wraps(fn)
        def _minimize(indv):
            return -fn(indv)
        return _minimize




#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Flip Bit mutation implementation. '''

from random import random, uniform

from ...mpiutil import MPIUtil
from ...plugin_interfaces.operators.mutation import Mutation
from ...components.binary_individual import BinaryIndividual
from ...components.decimal_individual import DecimalIndividual

mpi = MPIUtil()


class FlipBitMutation(Mutation):
    ''' Mutation operator with Flip Bit mutation implementation.

    :param pm: The probability of mutation (usually between 0.001 ~ 0.1)
    :type pm: float in range (0.0, 1.0]
    '''
    def __init__(self, pm):
        if pm <= 0.0 or pm > 1.0:
            raise ValueError('Invalid mutation probability')

        self.pm = pm

    def mutate(self, individual, engine):
        ''' Mutate the individual.

        :param individual: The individual on which crossover operation occurs
        :type individual: :obj:`lsga.components.IndividualBase`

        :param engine: Current genetic algorithm engine
        :type engine: :obj:`lsga.engine.GAEngine`

        :return: A mutated individual
        :rtype: :obj:`lsga.components.IndividualBase`
        '''
        do_mutation = True if random() <= self.pm else False

        if do_mutation:
            for i, genome in enumerate(individual.chromsome):
                no_flip = True if random() > self.pm else False
                if no_flip:
                    continue

                if type(individual) is BinaryIndividual:
                    individual.chromsome[i] = genome^1
                elif type(individual) is DecimalIndividual:
                    a, b = individual.ranges[i]
                    eps = individual.precisions[i]
                    n_intervals = (b - a)//eps
                    n = int(uniform(0, n_intervals + 1))
                    individual.chromsome[i] = a + n*eps
                else:
                    raise TypeError('Wrong individual type: {}'.format(type(individual)))

            # Update solution.
            individual.solution = individual.decode()

        return individual


class FlipBitBigMutation(FlipBitMutation):
    ''' Mutation operator using Flip Bit mutation implementation with adaptive
    big mutation rate to overcome premature or local-best solution.

    :param pm: The probability of mutation (usually between 0.001 ~ 0.1)
    :type pm: float in (0.0, 1.0]

    :param pbm: The probability of big mutation, usually more than 5 times
                bigger than pm.
    :type pbm: float

    :param alpha: intensive factor
    :type alpha: float, in range (0.5, 1)
    '''
    def __init__(self, pm, pbm, alpha):
        super(self.__class__, self).__init__(pm)

        if not (0.0 < pbm < 1.0):
            raise ValueError('Invalid big mutation probability')
        if pbm < 5*pm and mpi.is_master:
            self.logger.warning('Relative low probability for big mutation')
        self.pbm = pbm

        # Intensive factor.
        if not (0.5 < alpha < 1.0):
            raise ValueError('Invalid intensive factor, should be in (0.5, 1.0)')
        self.alpha = alpha

    def mutate(self, individual, engine):
        ''' Mutate the individual with adaptive big mutation rate.

        :param individual: The individual on which crossover operation occurs
        :type individual: :obj:`lsga.components.IndividualBase`

        :param engine: Current genetic algorithm engine
        :type engine: :obj:`lsga.engine.GAEngine`

        :return: A mutated individual
        :rtype: :obj:`lsga.components.IndividualBase`
        '''
        pm = self.pm

        if engine.fmax*self.alpha < engine.fmean:
            self.pm = self.pbm

        # Mutate with big probability.
        individual = super(self.__class__, self).mutate(individual, engine)

        # Recover probability.
        self.pm = pm

        return individual




#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Module for all test in ga.py
'''

import unittest

from .individual_test import IndividualTest
from .population_test import PopulationTest
from .roulette_wheel_selection_test import RouletteWheelSelectionTest
from .uniform_crossover_test import UniformCrossoverTest
from .flip_bit_mutation_test import FlipBitMutationTest
from .mpiutil_test import MPIUtilTest
from .engine_test import GAEngineTest
from .tournament_selection_test import TournamentSelectionTest
from .linear_ranking_selection_test import LinearRankingSelectionTest
from .exponential_ranking_selection_test import ExponentialRankingSelectionTest
from .linear_scaling_test import LinearScalingTest
from .dynamic_linear_scaling_test import DynamicLinearScalingTest
from .flip_bit_big_mutation_test import FlipBitBigMutationTest

def suite():
    ''' Generate test suite for all test cases in GAFT
    '''
    test_cases = [
        IndividualTest,
        PopulationTest,
        RouletteWheelSelectionTest,
        UniformCrossoverTest,
        FlipBitMutationTest,
        MPIUtilTest,
        GAEngineTest,
        TournamentSelectionTest,
        LinearRankingSelectionTest,
        ExponentialRankingSelectionTest,
        LinearScalingTest,
        DynamicLinearScalingTest,
        FlipBitBigMutationTest
    ]

    test_suite = unittest.TestSuite([
        unittest.TestLoader().loadTestsFromTestCase(tc) for tc in test_cases
    ])

    return test_suite

if '__main__' == __name__:
    result = unittest.TextTestRunner(verbosity=2).run(suite())

    if result.errors or result.failures:
        raise ValueError('Get erros and failures')




#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
A high-level utility class for parallelizing Genetic Algorithm by using MPI interfaces
in distributed MPI environment.
'''

import logging
from itertools import chain
from functools import wraps

try:
    from mpi4py import MPI
    MPI_INSTALLED = True
except ImportError:
    MPI_INSTALLED = False


class Singleton(type):
    def __call__(cls, *args, **kwargs):
        if not hasattr(cls, '_instance'):
            cls._instance = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instance


class MPIUtil(object):

    __metaclass__ = Singleton

    def __init__(self):
        ''' Wrapper class for higher level of MPI interfaces that will create a
        singleton for parallelization.
        '''
        logger_name = 'lsga.{}'.format(self.__class__.__name__)
        self._logger = logging.getLogger(logger_name)

    def bcast(self, data):
        ''' Broadcast data to MPI processes

        :param data: Data to be broadcasted
        :type data: any Python object
        '''
        if MPI_INSTALLED:
            mpi_comm = MPI.COMM_WORLD
            bdata = mpi_comm.bcast(data, root=0)
        else:
            bdata = data
        return bdata

    # Wrapper for common MPI interfaces.
    def barrier(self):
        ''' Block until all processes in the communicator have reached this routine
        '''
        if MPI_INSTALLED:
            mpi_comm = MPI.COMM_WORLD
            mpi_comm.barrier()

    @property
    def rank(self):
        ''' Get the rank of the calling process in the communicator
        '''
        if MPI_INSTALLED:
            mpi_comm = MPI.COMM_WORLD
            return mpi_comm.Get_rank()
        else:
            return 0

    @property
    def size(self):
        ''' Get the size of the group associated with a communicator
        '''
        if MPI_INSTALLED:
            mpi_comm = MPI.COMM_WORLD
            return mpi_comm.Get_size()
        else:
            return 1

    @property
    def is_master(self):
        ''' If current process is the master process
        '''
        return self.rank == 0

    # Utility methods.
    def split_seq(self, sequence):
        ''' Split the sequence according to rank and processor number.

        :param sequence: Data sequence to be splitted
        :type sequence: any Python object list

        :return: Sub data sequence for current process
        :rtype: any Python object list
        '''
        starts = [i for i in range(0, len(sequence), len(sequence)//self.size)]
        ends = starts[1: ] + [len(sequence)]
        start, end = list(zip(starts, ends))[self.rank]

        return sequence[start: end]

    def split_size(self, size):
        ''' Split a size number(int) to sub-size number.

        :param size: The size number to be splitted.
        :type size: int

        :return: Sub-size for current process
        :rtype: int
        '''
        if size < self.size:
            warn_msg = ('Splitting size({}) is smaller than process ' +
                        'number({}), more processor would be ' +
                        'superflous').format(size, self.size)
            self._logger.warning(warn_msg)
            splited_sizes = [1]*size + [0]*(self.size - size)
        elif size % self.size != 0:
            residual = size % self.size
            splited_sizes = [size // self.size]*self.size
            for i in range(residual):
                splited_sizes[i] += 1
        else:
            splited_sizes = [size // self.size]*self.size

        return splited_sizes[self.rank]

    def merge_seq(self, seq):
        ''' Gather data in sub-process to root process.

        :param seq: Sub data sequence for current process
        :type seq: any Python object list

        :return: Merged data sequence from all processes in a communicator
        :rtype: any Python object list
        '''
        if self.size == 1:
            return seq

        mpi_comm = MPI.COMM_WORLD
        merged_seq= mpi_comm.allgather(seq)
        return list(chain(*merged_seq))


def master_only(func):
    ''' Decorator to limit a function to be called only in master process in MPI env.
    '''
    @wraps(func)
    def _call_in_master_proc(*args, **kwargs):
        mpi = MPIUtil()
        if mpi.is_master:
            return func(*args, **kwargs)

    return _call_in_master_proc




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .individual import IndividualBase
from joblib import Parallel, delayed


class Memoized(object):
    ''' Descriptor for population statistical varibles caching.
    '''

    def __init__(self, func):
        self.func = func
        self.result = None
        self.fitness = None

    def __get__(self, instance, cls):
        self.instance = instance
        return self

    def __call__(self, fitness):
        if ((not self.instance.updated)  # population not changed
                and (self.result is not None)  # result already cached
                and (fitness == self.fitness)):  # fitness not changed
            # Return cached result directly.
            return self.result
        else:
            # Update fitness function.
            self.fitness = fitness
            # Update and memoize result.
            self.result = self.func(self.instance, fitness)
            # Recover flag.
            self.instance._updated = False
            return self.result


class Individuals(object):
    ''' Descriptor for all individuals in population.

    .. Note::
        Use this descriptor to ensure the individual related flags can be updated
        when the population indivduals are changed.
    '''

    def __init__(self, name):
        self.name = '_{}'.format(name)

    def __get__(self, instance, owner):
        return instance.__dict__[self.name]

    def __set__(self, instance, value):
        instance.__dict__[self.name] = value
        # Update flag.
        instance.update_flag()


class Population(object):
    ''' Class for representing population in genetic algorithm.

    :param indv_template: A template individual to clone all the other
                          individuals in current population.
    :type indv_template: :obj:`lsga.components.IndividualBase`

    :param size: The size of population, number of individuals in population.
    :type size: int
    '''
    # All individuals.
    individuals = Individuals('individuals')

    def __init__(self, indv_template, n_jobs=1, size=100):
        self.n_jobs = n_jobs
        # Population size.
        if size % 2 != 0:
            raise ValueError('Population size must be an even number')
        self.size = size

        # Template individual.
        self.indv_template = indv_template

        # Flag for monitoring changes of population.
        self._updated = False

        # Container for all individuals.
        class IndvList(list):
            ''' A proxy class inherited from built-in list to contain all
            individuals which can update the population._updated flag
            automatically when its content is changed.
            '''

            # {{{
            # NOTE: Use 'this' here to avoid name conflict.
            def __init__(this, *args):
                super(this.__class__, this).__init__(*args)

            def __setitem__(this, key, value):
                '''
                Override __setitem__ in built-in list type.
                '''
                old_value = this[key]
                if old_value == value:
                    return
                super(this.__class__, self).__setitem__(key, value)
                # Update population flag.
                self.update_flag()

            def append(this, item):
                '''
                Override append method of built-in list type.
                '''
                super(this.__class__, this).append(item)
                # Update population flag.
                self.update_flag()

            def extend(this, iterable_item):
                if not iterable_item:
                    return
                super(this.__class__, this).extend(iterable_item)
                # Update population flag.
                self.update_flag()
            # }}}

        self._individuals = IndvList()

    def init(self, indvs=None):
        ''' Initialize current population with individuals.

        :param indvs: Initial individuals in population, randomly initialized
                      individuals are created if not provided.
        :type indvs: list of Individual object
        '''
        IndvType = self.indv_template.__class__

        if indvs is None:
            for _ in range(self.size):
                indv = IndvType(ranges=self.indv_template.ranges,
                                eps=self.indv_template.eps)
                self.individuals.append(indv)
        else:
            # Check individuals.
            if len(indvs) != self.size:
                raise ValueError('Invalid individuals number')
            for indv in indvs:
                if not isinstance(indv, IndividualBase):
                    raise ValueError('individual class must be subclass of IndividualBase')
            self.individuals = indvs

        self._updated = True

        return self

    def update_flag(self):
        ''' Interface for updating individual update flag to True.
        '''
        self._updated = True

    @property
    def updated(self):
        ''' Query function for population updating flag.
        '''
        return self._updated

    def new(self):
        ''' Create a new emtpy population.
        '''
        return self.__class__(indv_template=self.indv_template,
                              size=self.size)

    def __getitem__(self, key):
        '''
        Get individual by index.
        '''
        if key < 0 or key >= self.size:
            raise IndexError('Individual index({}) out of range'.format(key))
        return self.individuals[key]

    def __len__(self):
        '''
        Get length of population.
        '''
        return len(self.individuals)

    def best_indv(self, fitness):
        ''' The individual with the best fitness.

        :param fitness: Fitness function to calculate fitness value
        :type fitness: function

        :return: the best individual in current population
        :rtype: :obj:`lsga.components.IndividualBase`
        '''

        all_fits = self.all_fits(fitness)
        return max(self.individuals,
                   key=lambda indv: all_fits[self.individuals.index(indv)])

    def worst_indv(self, fitness):
        ''' The individual with the worst fitness.

        :param fitness: Fitness function to calculate fitness value
        :type fitness: function

        :return: the worst individual in current population
        :rtype: :obj:`lsga.components.IndividualBase`
        '''
        all_fits = self.all_fits(fitness)
        return min(self.individuals,
                   key=lambda indv: all_fits[self.individuals.index(indv)])

    def max(self, fitness):
        ''' Get the maximum fitness value in population.

        :param fitness: Fitness function to calculate fitness value
        :type fitness: function

        :return: The maximum fitness value
        :rtype: float
        '''
        return max(self.all_fits(fitness))

    def min(self, fitness):
        ''' Get the minimum value of fitness in population.

        :param fitness: Fitness function to calculate fitness value
        :type fitness: function

        :return: The minimum fitness value
        :rtype: float
        '''
        return min(self.all_fits(fitness))

    def mean(self, fitness):
        ''' Get the average fitness value in population.

        :param fitness: Fitness function to calculate fitness value
        :type fitness: function

        :return: The average fitness value
        :rtype: float
        '''
        all_fits = self.all_fits(fitness)
        return sum(all_fits) / len(all_fits)

    @Memoized
    def all_fits(self, fitness):
        ''' Get all fitness values in population.

        :param fitness: Fitness function to calculate fitness value
        :type fitness: function
        '''

        return Parallel(n_jobs=self.n_jobs, verbose=10)(delayed(fitness)(indv) for indv in self.individuals)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from lsga import GAEngine
from lsga.components import *
from lsga.operators import *
from lsga.analysis.fitness_store import FitnessStore
from lsga.analysis.console_output import ConsoleOutput


class EngineWrapper():

    def create_engine(self, ranges, n_jobs=1, eps=0.001, population_size=50, encoding='binary', selection='tournament',
                      crossover='uniform', mutation='flip_bit'):
        """

        :param ranges:
        :param n_jobs:
        :param eps:
        :param population_size:
        :param encoding:
        :param selection:
        :param crossover:
        :param mutation:
        """

        assert encoding in ['binary', 'decimal']
        assert selection in ['tournament', 'roulette_wheel', 'linear_ranking', 'exponential_ranking']
        assert crossover in ['uniform']
        assert mutation in ['flip_bit']

        if encoding == 'binary':
            indv_template = BinaryIndividual(ranges=ranges, eps=eps)
        else:
            indv_template = BinaryIndividual(ranges=ranges, eps=eps)

        if selection == 'tournament':
            lsga_selection = TournamentSelection()
        elif selection == 'roulette_wheel':
            lsga_selection = RouletteWheelSelection()
        elif selection == 'linear_ranking':
            lsga_selection = LinearRankingSelection()
        else:
            lsga_selection = ExponentialRankingSelection()

        if crossover == 'uniform':
            lsga_crossover = UniformCrossover(pc=0.8, pe=0.5)

        if mutation == 'flip_bit':
            lsga_mutation = FlipBitBigMutation(pm=0.1, pbm=0.55, alpha=0.6)

        population = Population(indv_template=indv_template, size=population_size, n_jobs=n_jobs).init()

        engine = GAEngine(population=population, selection=lsga_selection,
                          crossover=lsga_crossover, mutation=lsga_mutation,
                          analysis=[ConsoleOutput, FitnessStore])

        return engine



#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Tournament Selection implementation. '''

from random import sample

from ...plugin_interfaces.operators.selection import Selection


class TournamentSelection(Selection):
    ''' Selection operator using Tournament Strategy with tournament size equals
    to two by default.

    :param tournament_size: Individual number in one tournament
    :type tournament_size: int
    '''
    def __init__(self, tournament_size=2):
        self.tournament_size = tournament_size

    def select(self, population, fitness):
        ''' Select a pair of parent using Tournament strategy.

        :param population: Population where the selection operation occurs.
        :type population: :obj:`lsga.components.Population`

        :return: Selected parents (a father and a mother)
        :rtype: list of :obj:`lsga.components.IndividualBase`
        '''
        # Competition function.
        all_fits = population.all_fits(fitness)
        def complete(competitors):
            '''
            Competition function.
            '''
            key = lambda indv: all_fits[population.individuals.index(indv)]
            return max(competitors, key=key)

        # Check validity of tournament size.
        if self.tournament_size >= len(population):
            msg = 'Tournament size({}) is larger than population size({})'
            raise ValueError(msg.format(self.tournament_size, len(population)))

        # Pick winners of two groups as parent.
        competitors_1 = sample(population.individuals, self.tournament_size)
        competitors_2 = sample(population.individuals, self.tournament_size)
        father, mother = complete(competitors_1), complete(competitors_2)

        return father, mother




import logging
import sys

from .engine import GAEngine
from .engine_wrapper import EngineWrapper

__version__ = '0.5.7'
__author__ = 'ShaoZhengjiang <shaozhengjiang@gmail.com>'

# Set root logger.
logger = logging.getLogger('lsga')
logger.setLevel(logging.INFO)
console_hdlr = logging.StreamHandler(sys.stdout)
console_hdlr.setLevel(logging.INFO)
formatter = logging.Formatter('%(name)s   %(levelname)-8s %(message)s')
console_hdlr.setFormatter(formatter)
logger.addHandler(console_hdlr)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Exponential Ranking Selection implemention. '''

from random import random
from itertools import accumulate
from bisect import bisect_right

from ...plugin_interfaces.operators.selection import Selection


class ExponentialRankingSelection(Selection): 
    ''' Selection operator using Exponential Ranking selection method.

    :param base: The base of exponent
    :type base: float in range (0.0, 1.0)
    '''
    def __init__(self, base=0.5):
        if not (0.0 < base < 1.0):
            raise ValueError('The base of exponent c must in range (0.0, 1.0)')

        self.base = base

    def select(self, population, fitness):
        ''' Select a pair of parent individuals using exponential ranking method.

        :param population: Population where the selection operation occurs.
        :type population: :obj:`lsga.components.Population`

        :return: Selected parents (a father and a mother)
        :rtype: list of :obj:`lsga.components.IndividualBase`
        '''
        # Individual number.
        NP = len(population)

        # Add rank to all individuals in population.
        all_fits = population.all_fits(fitness)
        indvs = population.individuals
        sorted_indvs = sorted(indvs,
                              key=lambda indv: all_fits[indvs.index(indv)])

        # NOTE: Here the rank i belongs to {1, ..., N}
        p = lambda i: self.base**(NP - i)
        probabilities = [p(i) for i in range(1, NP + 1)]

        # Normalize probabilities.
        psum = sum(probabilities)
        wheel = list(accumulate([p/psum for p in probabilities]))

        # Select parents.
        father_idx = bisect_right(wheel, random())
        father = sorted_indvs[father_idx]
        mother_idx = (father_idx + 1) % len(wheel)
        mother = sorted_indvs[mother_idx]

        return father, mother




#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Module for Genetic Algorithm selection operator class '''

from ..metaclasses import SelectionMeta

class Selection(metaclass=SelectionMeta):
    ''' Class for providing an interface to easily extend the behavior of selection
    operation.
    '''

    def select(self, population, fitness):
        ''' Called when we need to select parents from a population to later breeding.

        :param population: The current population
        :type population: lsga.compoenents.Population

        :return parents: Two selected individuals for crossover
        :type parents: tuple of gaft.components.IndividualBase
        '''
        raise NotImplementedError




#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Test case for Population
'''

import unittest

from lsga.components import Population, BinaryIndividual

class PopulationTest(unittest.TestCase):

    def setUp(self):
        self.maxDiff = True
        self.indv_template = BinaryIndividual(ranges=[(0, 1)])
        def fitness(indv):
            x, = indv.solution
            return x**3 - 60*x**2 + 900*x + 100
        self.fitness = fitness

    def test_initialization(self):
        ''' Make sure a population can be initialized correctly. '''
        population = Population(indv_template=self.indv_template, size=10)

        self.assertListEqual(population.individuals, [])

        population.init()
        self.assertEqual(len(population.individuals), 10)

        # Check individual.
        self.assertTrue(isinstance(population[0], BinaryIndividual))

    def test_new_population(self):
        ''' Make sure population can clone a new population. '''
        population = Population(indv_template=self.indv_template, size=10)
        population.init()
        new_population = population.new()
        self.assertEqual(new_population.size, 10)
        self.assertListEqual(new_population.individuals, [])

    def test_all_fits(self):
        population = Population(indv_template=self.indv_template, size=10)
        population.init()
        all_fits = population.all_fits(fitness=self.fitness)

        self.assertEqual(len(all_fits), 10)

        for fit in all_fits:
            self.assertTrue(type(fit) is float)

if '__main__' == __name__:
    suite = unittest.TestLoader().loadTestsFromTestCase(PopulationTest)
    unittest.TextTestRunner(verbosity=2).run(suite)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Module for Genetic Algorithm mutation operator class '''

from ..metaclasses import MutationMeta


class Mutation(metaclass=MutationMeta):
    ''' Class for providing an interface to easily extend the behavior of selection
    operation.

    Attributes:

        pm(float): Default mutation probability, default is 0.1
    '''
    # Default mutation probability.
    pm = 0.1

    def mutate(self, individual, engine):
        ''' Called when an individual to be mutated.

        :param individual: The individual to be mutated
        :type individual: gaft.components.IndividualBase

        :param engine: The GA engine where the mutation operator belongs.
        :type engine: gaft.engine.GAEngine
        '''
        raise NotImplementedError




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from ..plugin_interfaces.analysis import OnTheFlyAnalysis


class ConsoleOutput(OnTheFlyAnalysis):
    ''' Built-in on-the-fly analysis plugin class for outputing log on console.

    Attribute:

        interval(:obj:`int`): The analysis interval in evolution iteration, default 
                              value is 1 meaning analyze every step.

        master_only(:obj:`bool`): Flag for if the analysis plugin is only effective 
                                  in master process. Default is True.
    '''

    # Analysis interval.
    interval = 1

    # Only analyze in master process?
    master_only = True

    def setup(self, ng, engine):
        generation_info = 'Generation number: {}'.format(ng)
        population_info = 'Population number: {}'.format(engine.population.size)
        self.logger.info('{} {}'.format(generation_info, population_info))

    def register_step(self, g, population, engine):
        best_indv = population.best_indv(engine.fitness)
        ng_info = 'Generation: {}, '.format(g+1)
        fit_info = 'best fitness: {:.3f}, '.format(engine.ori_fmax)
        scaled_info = 'scaled fitness: {:.3f}'.format(engine.fmax)
        msg = ng_info + fit_info + scaled_info
        self.logger.info(msg)

    def finalize(self, population, engine):
        best_indv = population.best_indv(engine.fitness)
        x = best_indv.solution
        y = engine.ori_fmax
        msg = 'Optimal solution: ({}, {})'.format(x, y)
        self.logger.info(msg)




''' Package for built-in selection operators '''

from .exponential_ranking_selection import ExponentialRankingSelection
from .linear_ranking_selection import LinearRankingSelection
from .roulette_wheel_selection import RouletteWheelSelection
from .tournament_selection import TournamentSelection




#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Module for Genetic Algorithm crossover operator class '''

from ..metaclasses import CrossoverMeta


class Crossover(metaclass=CrossoverMeta):
    ''' Class for providing an interface to easily extend the behavior of crossover
    operation between two individuals for children breeding.

    Attributes:

        pc(:obj:`float`): The probability of crossover (usaully between 0.25 ~ 1.0)
    '''

    # The probability of crossover (usaully between 0.25 ~ 1.0)
    pc = 0.8

    def cross(self, father, mother):
        ''' Called when we need to cross parents to generate children.

        :param father: The parent individual to be crossed
        :type father: gaft.components.IndividualBase

        :param mother: The parent individual to be crossed
        :type mother: gaft.components.IndividualBase

        :return children: Two new children individuals
        :type children: tuple of gaft.components.IndividualBase
        '''
        raise NotImplementedError




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from ..plugin_interfaces.analysis import OnTheFlyAnalysis

class FitnessStore(OnTheFlyAnalysis):
    ''' Built-in on-the-fly analysis plugin class for storing fitness related data during iteration.

    Attribute:
        interval(:obj:`int`): The analysis interval in evolution iteration, default 
                              value is 1 meaning analyze every step.
        master_only(:obj:`bool`): Flag for if the analysis plugin is only effective 
                                  in master process. Default is True.
    '''

    # Analysis interval.
    interval = 1

    # Only analyze in master process?
    master_only = True

    def setup(self, ng, engine):
        # Generation numbers.
        self.ngs = []

        # Best fitness in each generation.
        self.fitness_values = []

        # Best solution.
        self.solution = []

    def register_step(self, g, population, engine):
        # Collect data.
        best_indv = population.best_indv(engine.fitness)
        best_fit = engine.ori_fmax

        self.ngs.append(g)
        self.solution.append(best_indv.solution)
        self.fitness_values.append(best_fit)

    def finalize(self, population, engine):
        with open('best_fit.py', 'w', encoding='utf-8') as f:
            f.write('best_fit = [\n')
            for ng, x, y in zip(self.ngs, self.solution, self.fitness_values):
                f.write('    ({}, {}, {}),\n'.format(ng, x, y))
            f.write(']\n\n')

        self.logger.info('Best fitness values are written to best_fit.py')




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import inspect
from functools import wraps

from ..components.individual import IndividualBase
from ..components.population import Population
from ..mpiutil import master_only


class AnalysisMeta(type):
    ''' Metaclass for analysis plugin class
    '''
    def __new__(cls, name, bases, attrs):
        # Check interval type.
        if 'interval' in attrs:
            interval = attrs['interval']
            if type(interval) is not int or interval <= 0:
                raise TypeError('analysis interval must be a positive integer')

        for method_name in ['setup', 'register_step', 'finalize']:
            method = attrs.get(method_name, None)
            if method is not None and not callable(method):
                msg = "{} must be a callable object".format(method)
                raise AttributeError(msg)
            # Set default interface methods.
            elif method is None:
                if method_name == 'setup':
                    attrs[method_name] = lambda self, ng, engine: None
                elif method_name == 'register_step':
                    attrs[method_name] = lambda self, g, population, engine: None
                elif method_name == 'finalize':
                    attrs[method_name] = lambda self, population, engine: None

        # Check if the plugin is only used in master process.
        called_in_master = attrs['master_only'] if 'master_only' in attrs else False

        # Wrap all interfaces.
        if called_in_master:
            for method_name in ['setup', 'register_step', 'finalize']:
                attrs[method_name] = master_only(attrs[method_name])

        # Set logger.
        logger_name = 'lsga.{}'.format(name)
        attrs['logger'] = logging.getLogger(logger_name)

        return type.__new__(cls, name, bases, attrs)


class CrossoverMeta(type):
    ''' Metaclass for crossover operator class.
    '''
    def __new__(cls, name, bases, attrs):
        if 'cross' not in attrs:
            raise AttributeError('crossover operator class must have cross method')

        if 'pc' in attrs and (attrs['pc'] <= 0.0 or attrs['pc'] > 1.0):
            raise ValueError('Invalid crossover probability')

        cross = attrs['cross']

        # Check parameter of cross method.
        sig = inspect.signature(cross)
        if 'father' not in sig.parameters:
            raise NameError('cross method must have father parameter')
        if 'mother' not in sig.parameters:
            raise NameError('cross method must have mother parameter')

        # Add parameter check to user-defined method.
        @wraps(cross)
        def _wrapped_cross(self, father, mother):
            ''' Wrapper to add parameters type checking.
            '''
            # Check parameter types.
            if not (isinstance(father, IndividualBase) and
                    isinstance(mother, IndividualBase)):
                raise TypeError('father and mother\'s type must be subclass of IndividualBase')

            return cross(self, father, mother)

        attrs['cross'] = _wrapped_cross

        # Set logger.
        logger_name = 'lsga.{}'.format(name)
        attrs['logger'] = logging.getLogger(logger_name)

        return type.__new__(cls, name, bases, attrs)


class MutationMeta(type):
    ''' Metaclass for mutation operator class.
    '''
    def __new__(cls, name, bases, attrs):
        if 'mutate' not in attrs:
            raise AttributeError('mutation operator class must have mutate method')

        if 'pm' in attrs and (attrs['pm'] <= 0.0 or attrs['pm'] > 1.0):
            raise ValueError('Invalid mutation probability')

        mutate = attrs['mutate']

        # Check parameters of mutate method.
        sig = inspect.signature(mutate)
        if 'individual' not in sig.parameters:
            raise NameError('mutate method must have individual parameter')

        # Add parameter check to user-defined method.
        @wraps(mutate)
        def _wrapped_mutate(self, individual, engine):
            ''' Wrapper to add parameters type checking.
            '''
            # Check parameter types.
            if not isinstance(individual, IndividualBase):
                raise TypeError('individual\' type must be subclass of IndividualBase')

            return mutate(self, individual, engine)

        attrs['mutate'] = _wrapped_mutate

        # Set logger.
        logger_name = 'lsga.{}'.format(name)
        attrs['logger'] = logging.getLogger(logger_name)

        return type.__new__(cls, name, bases, attrs)


class SelectionMeta(type):
    ''' Metaclass for selection operator class.
    '''
    def __new__(cls, name, bases, attrs):
        # Check select method.
        if 'select' not in attrs:
            raise AttributeError('selection operator class must have select method')

        select = attrs['select']

        # Check select arguments.
        sig = inspect.signature(select)
        if 'population' not in sig.parameters:
            raise NameError('select method must have population parameter')
        if 'fitness' not in sig.parameters:
            raise NameError('select method must have fitness parameter')

        # Add parameter check to user-defined method.
        @wraps(select)
        def _wrapped_select(self, population, fitness):
            ''' Wrapper to add parameters type checking.
            '''
            # Check parameter types.
            if not isinstance(population, Population):
                raise TypeError('population must be Population object')
            if not callable(fitness):
                raise TypeError('fitness must be a callable object')

            return select(self, population, fitness)

        attrs['select'] = _wrapped_select

        # Set logger.
        logger_name = 'lsga.{}'.format(name)
        attrs['logger'] = logging.getLogger(logger_name)

        return type.__new__(cls, name, bases, attrs)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Test case for built-in Flip Bit mutation operator with big mutation rate.
'''

import unittest
from math import sin, cos

from .. import GAEngine
from ..components import BinaryIndividual
from ..components import Population
from ..operators import RouletteWheelSelection
from ..operators import UniformCrossover
from ..operators.mutation.flip_bit_mutation import FlipBitBigMutation

class FlipBitBigMutationTest(unittest.TestCase):

    def setUp(self):
        self.maxDiff

    def test_mutate(self):
        ''' Make sure the individual can be mutated correctly.
        '''
        indv_template = BinaryIndividual(ranges=[(0, 10)], eps=0.001)
        population = Population(indv_template=indv_template, size=50).init()

        # Create genetic operators.
        selection = RouletteWheelSelection()
        crossover = UniformCrossover(pc=0.8, pe=0.5)
        mutation = FlipBitBigMutation(pm=0.03, pbm=0.2, alpha=0.6)

        # Create genetic algorithm engine.
        engine = GAEngine(population=population, selection=selection,
                          crossover=crossover, mutation=mutation)

        @engine.fitness_register
        def fitness(indv):
            x, = indv.solution
            return x + 10*sin(5*x) + 7*cos(4*x)

        mutation.mutate(indv_template, engine)

if '__main__' == __name__:
    suite = unittest.TestLoader().loadTestsFromTestCase(FlipBitBigMutationTest)
    unittest.TextTestRunner(verbosity=2).run(suite)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Test the linear scaling decorator of GA engine.
'''

import unittest
from math import sin, cos

from lsga import GAEngine
from lsga.components import BinaryIndividual
from lsga.components import Population
from lsga.operators import RouletteWheelSelection
from lsga.operators import UniformCrossover
from lsga.operators import FlipBitMutation


class LinearScalingTest(unittest.TestCase):

    def setUp(self):
        self.maxDiff = True

    def test_run(self):
        '''
        Make sure GA engine can run correctly.
        '''
        indv_template = BinaryIndividual(ranges=[(0, 10)], eps=0.001)
        population = Population(indv_template=indv_template, size=50).init()

        # Create genetic operators.
        selection = RouletteWheelSelection()
        crossover = UniformCrossover(pc=0.8, pe=0.5)
        mutation = FlipBitMutation(pm=0.1)

        # Create genetic algorithm engine.
        engine = GAEngine(population=population, selection=selection,
                          crossover=crossover, mutation=mutation)

        @engine.fitness_register
        @engine.linear_scaling(target='max', ksi=0.5)
        def fitness(indv):
            x, = indv.solution
            return x + 10*sin(5*x) + 7*cos(4*x)

        engine.run(50)

if '__main__' == __name__:
    suite = unittest.TestLoader().loadTestsFromTestCase(LinearScalingTest)
    unittest.TextTestRunner(verbosity=2).run(suite)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from lsga import __version__ as version

maintainer = 'Marko Stojanovic'
maintainer_email = 'marko.stojanovickg@outlook.com  '
author = maintainer
author_email = maintainer_email
description = "A Genetic Algorithm Framework in Python"
long_description = '''
====
LSGA Package
====


'''

install_requires = [
]

license = 'LICENSE'

name = 'lsga'
platforms = ['linux', 'windows', 'macos']
url = 'https://github.com/mstojanovickg/lsga'
download_url = 'https://github.com/mstojanovickg/lsga/releases'

classifiers = [
    'Development Status :: 3 - Alpha',
    'Topic :: Utilities',
    'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)'
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
]

test_suite = 'lsga.tests.test_all'

setup(
    author=author,
    author_email=author_email,
    description=description,
    license=license,
    long_description=long_description,
    install_requires=install_requires,
    maintainer=maintainer,
    name=name,
    packages=find_packages(),
    platforms=platforms,
    url=url,
    download_url=download_url,
    version=version,
    test_suite=test_suite
)

